table.CALSTABLE > tbody > tr:nth-child(1) > td:nth-child(1) {
    width: 20em;
}
.synopsis, .classsynopsis {
    background: #eeeeee;
    border: solid 1px #aaaaaa;
}
.programlisting {
    background: #eeeeee;
    border: solid 1px #000000;
}
.navigation {
    background: #eeeeee;
    border: solid 1px #000000;
}
.navigation a {
    color: initial;
}
.navigation a:visited {
    color: initial;
}

h1 {
    text-align: center !important;
}

#xmlsecContent {
    padding-left: 8em;
}

#xmlsecContent table {
    border-collapse: collapse;
    width: 100%;
}

#xmlsecContent th, #xmlsecContent td {
  border: 1px solid black;
  padding: 8px;
}

#xmlsecContent td.supported-feature {
  background-color: #C1E1C1; /* Light green */
}