/**
 * @file wincrypt.h
 * @brief Stub wincrypt.h for doxygen documentation generation on non-Windows platforms.
 */
#ifndef __XMLSEC_DOXYGEN_WINCRYPT_H__
#define __XMLSEC_DOXYGEN_WINCRYPT_H__
/* All types covered by windows.h stub */
#endif /* __XMLSEC_DOXYGEN_WINCRYPT_H__ */
