/**
 * version.rc:
 *
 * Windows resource file for xmlsec binaries. Compiled with /D defines
 * to set per-target values:
 *
 *   RC_VERSION_MAJOR, RC_VERSION_MINOR, RC_VERSION_MICRO  - version numbers
 *   RC_FILENAME      - original file name (e.g. "libxmlsec.dll")
 *   RC_DESCRIPTION   - file description string
 *   RC_FILETYPE      - VFT_DLL or VFT_APP
 */

#include <winver.h>

#ifndef RC_VERSION_MAJOR
#define RC_VERSION_MAJOR 0
#endif
#ifndef RC_VERSION_MINOR
#define RC_VERSION_MINOR 0
#endif
#ifndef RC_VERSION_MICRO
#define RC_VERSION_MICRO 0
#endif
#ifndef RC_FILENAME
#define RC_FILENAME "xmlsec"
#endif
#ifndef RC_DESCRIPTION
#define RC_DESCRIPTION "XML Security Library"
#endif
#ifndef RC_FILETYPE
#define RC_FILETYPE VFT_DLL
#endif

#define RC_STRINGIFY2(x) #x
#define RC_STRINGIFY(x)  RC_STRINGIFY2(x)
#define RC_VERSION_STR   RC_STRINGIFY(RC_VERSION_MAJOR) "." RC_STRINGIFY(RC_VERSION_MINOR) "." RC_STRINGIFY(RC_VERSION_MICRO)

VS_VERSION_INFO VERSIONINFO
  FILEVERSION    RC_VERSION_MAJOR,RC_VERSION_MINOR,RC_VERSION_MICRO,0
  PRODUCTVERSION RC_VERSION_MAJOR,RC_VERSION_MINOR,RC_VERSION_MICRO,0
  FILEFLAGSMASK  VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
  FILEFLAGS      VS_FF_DEBUG
#else
  FILEFLAGS      0x0L
#endif
  FILEOS         VOS_NT_WINDOWS32
  FILETYPE       RC_FILETYPE
  FILESUBTYPE    VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "CompanyName",      "XMLSec Library\0"
            VALUE "FileDescription",  RC_DESCRIPTION "\0"
            VALUE "FileVersion",      RC_VERSION_STR "\0"
            VALUE "InternalName",     RC_FILENAME "\0"
            VALUE "LegalCopyright",   "Copyright (C) 2002-2026 Aleksey Sanin. All Rights Reserved.\0"
            VALUE "OriginalFilename", RC_FILENAME "\0"
            VALUE "ProductName",      "XML Security Library (xmlsec)\0"
            VALUE "ProductVersion",   RC_VERSION_STR "\0"
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END
